#ifndef _M68K_INIT_H
#define _M68K_INIT_H

#include <linux/config.h>

#ifndef CONFIG_KGDB

#define __init __attribute__ ((__section__ (".text.init")))
#define __initdata __attribute__ ((__section__ (".data.init")))
/* For assembly routines */
#define __INIT		.section	".text.init",#alloc,#execinstr
#define __FINIT		.previous
#define __INITDATA	.section	".data.init",#alloc,#write

#else

/* gdb doesn't like it all if the code for one source file isn't together in
 * the executable, so we must avoid the .init sections :-( */
	
#define __init
#define __initdata
/* For assembly routines */
#define __INIT
#define __FINIT
#define __INITDATA

#endif /* CONFIG_KGDB */

#endif
