int usb_acm_init(void);
int usb_audio_init(void);
int usb_hub_init(void);
void usb_hub_cleanup(void);
int usb_kbd_init(void);
void usb_major_init(void);
void usb_major_cleanup(void);
void usb_mouse_cleanup(void);
int usb_hp_scanner_init(void);
void usb_hp_scanner_cleanup(void);
int usb_printer_init(void);
int proc_usb_init (void);
void proc_usb_cleanup (void);
int usb_scsi_init(void);
int usb_serial_init (void);
