/*
 * linux/include/asm-arm/proc-armv/page.h
 *
 * Copyright (C) 1995, 1996 Russell King
 */
#ifndef __ASM_PROC_PAGE_H
#define __ASM_PROC_PAGE_H

/* PAGE_SHIFT determines the page size */
#define PAGE_SHIFT	12

#define EXEC_PAGESIZE   4096

#endif /* __ASM_PROC_PAGE_H */
