/* $Id: shmparam.h,v 1.4 1998/09/28 07:15:01 jj Exp $ */
#ifndef _ASMSPARC_SHMPARAM_H
#define _ASMSPARC_SHMPARAM_H

#define	SHMLBA PAGE_SIZE		 /* attach addr a multiple of this */

#endif /* _ASMSPARC_SHMPARAM_H */
