/*
 *  linux/include/asm-alpha/hdreg.h
 *
 *  Copyright (C) 1994-1996  Linus Torvalds & authors
 */

#ifndef __ASMalpha_HDREG_H
#define __ASMalpha_HDREG_H

typedef unsigned short ide_ioreg_t;

#endif /* __ASMalpha_HDREG_H */
