#ifndef _FLASH_H
#define _FLASH_H

#define FLASH_MINOR		 160	/* MAJOR is 10 - miscdevice */
#define CMD_WRITE_DISABLE	 0
#define CMD_WRITE_ENABLE	 0x28
#define CMD_WRITE_BASE64K_ENABLE 0x47

#endif /* _FLASH_H */
