/*
 *  linux/include/asm-m68k/hdreg.h
 *
 *  Copyright (C) 1994-1996  Linus Torvalds & authors
 */

#ifndef _M68K_HDREG_H
#define _M68K_HDREG_H

typedef unsigned int   q40ide_ioreg_t;
typedef unsigned char * ide_ioreg_t;

#endif /* _M68K_HDREG_H */
